@@ -1,13 +1,17 @@
 # -*- coding: utf-8 -*-
+
 """Python 3 API wrapper for Garmin Connect to get your statistics."""
+
 import json
 import logging
 import re
+import requests
 from enum import Enum, auto
 from typing import Any, Dict
 
 import cloudscraper
 
+
 logger = logging.getLogger(__name__)
 
 
@@ -15,6 +19,7 @@ class ApiClient:
     """Class for a single API endpoint."""
 
     default_headers = {
+        # 'User-agent': 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/535.2 (KHTML, like Gecko) Chrome/15.0.874.121 Safari/535.2'
         "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:66.0) Gecko/20100101 Firefox/66.0"
     }
 
@@ -31,6 +36,16 @@ class ApiClient:
         if aditional_headers:
             self.headers.update(aditional_headers)
 
+    def set_cookies(self, cookies):
+        logger.debug("Restoring cookies for saved session")
+        self.session.cookies.update(cookies)
+
+    def get_cookies(self):
+        return self.session.cookies
+
+    def clear_cookies(self):
+        self.session.cookies.clear()
+
     def url(self, addurl=None):
         """Return the url for the API endpoint."""
 
@@ -53,7 +68,7 @@ class ApiClient:
         try:
             response = self.session.get(url, headers=total_headers, params=params)
             response.raise_for_status()
-            logger.debug("Response: %s", response.content)
+            # logger.debug("Response: %s", response.content)
             return response
         except Exception as err:
             logger.debug("Response in exception: %s", response.content)
@@ -82,7 +97,7 @@ class ApiClient:
                 url, headers=total_headers, params=params, data=data
             )
             response.raise_for_status()
-            logger.debug("Response: %s", response.content)
+            # logger.debug("Response: %s", response.content)
             return response
         except Exception as err:
             logger.debug("Response in exception: %s", response.content)
@@ -99,8 +114,9 @@ class ApiClient:
 class Garmin:
     """Class for fetching data from Garmin Connect."""
 
-    def __init__(self, email, password, is_cn=False):
+    def __init__(self, email=None, password=None, is_cn=False, session_data=None):
         """Create a new class instance."""
+        self.session_data = session_data
 
         self.username = email
         self.password = password
@@ -128,17 +144,38 @@ class Garmin:
         self.garmin_connect_daily_summary_url = (
             "proxy/usersummary-service/usersummary/daily"
         )
-        self.garmin_connect_metrics_url = "proxy/metrics-service/metrics/maxmet/latest"
+        self.garmin_connect_metrics_url = "proxy/metrics-service/metrics/maxmet/daily"
         self.garmin_connect_daily_hydration_url = (
             "proxy/usersummary-service/usersummary/hydration/daily"
         )
         self.garmin_connect_personal_record_url = (
             "proxy/personalrecord-service/personalrecord/prs"
         )
-        self.garmin_connect_sleep_daily_url = (
+        self.garmin_connect_earned_badges_url = "proxy/badge-service/badge/earned"
+        self.garmin_connect_adhoc_challenges_url = (
+            "proxy/adhocchallenge-service/adHocChallenge/historical"
+        )
+        self.garmin_connect_badge_challenges_url = (
+            "proxy/badgechallenge-service/badgeChallenge/completed"
+        )
+        self.garmin_connect_available_badge_challenges_url = (
+            "proxy/badgechallenge-service/badgeChallenge/available"
+        )
+        self.garmin_connect_non_completed_badge_challenges_url = (
+            "proxy/badgechallenge-service/badgeChallenge/non-completed"
+        )
+        self.garmin_connect_daily_sleep_url = (
             "proxy/wellness-service/wellness/dailySleepData"
         )
-        self.garmin_connect_rhr = "proxy/userstats-service/wellness/daily"
+        self.garmin_connect_daily_stress_url = (
+            "proxy/wellness-service/wellness/dailyStress"
+        )
+
+        self.garmin_connect_rhr_url = "proxy/userstats-service/wellness/daily"
+
+        self.garmin_connect_training_readiness_url = "proxy/metrics-service/metrics/trainingreadiness"
+
+        self.garmin_connect_training_status_url = "proxy/metrics-service/metrics/trainingstatus/aggregated"
 
         self.garmin_connect_user_summary_chart = (
             "proxy/wellness-service/wellness/dailySummaryChart"
@@ -162,6 +199,9 @@ class Garmin:
         self.garmin_connect_gpx_download = "proxy/download-service/export/gpx/activity"
         self.garmin_connect_kml_download = "proxy/download-service/export/kml/activity"
         self.garmin_connect_csv_download = "proxy/download-service/export/csv/activity"
+
+        self.garmin_connect_fit_upload_url = "proxy/upload-service/upload/.fit"
+
         self.garmin_connect_gear = "proxy/gear-service/gear/filterGear"
 
         self.garmin_connect_logout = "auth/logout/?url="
@@ -196,9 +236,65 @@ class Garmin:
         return None
 
     def login(self):
+        if self.session_data is None:
+            return self.authenticate()
+        else:
+            return self.login_session()
+
+    def login_session(self):
+        logger.debug("Login with cookies")
+
+        session_display_name = self.session_data["display_name"]
+        logger.debug("Set cookies in session")
+        self.modern_rest_client.set_cookies(
+            requests.utils.cookiejar_from_dict(self.session_data["session_cookies"])
+        )
+        self.sso_rest_client.set_cookies(
+            requests.utils.cookiejar_from_dict(self.session_data["login_cookies"])
+        )
+
+        logger.debug("Get page data with cookies")
+        params = {
+            "service": "https://connect.garmin.com/modern/",
+            "webhost": "https://connect.garmin.com",
+            "gateway": "true",
+            "generateExtraServiceTicket": "true",
+            "generateTwoExtraServiceTickets": "true",
+        }
+        response = self.sso_rest_client.get("login", params=params)
+        logger.debug("Session response %s", response.status_code)
+        if response.status_code != 200:
+            logger.debug("Session expired, authenticating again!")
+            return self.authenticate()
+
+        user_prefs = self.__get_json(response.text, "VIEWER_USERPREFERENCES")
+        if user_prefs is None:
+            logger.debug("Session expired, authenticating again!")
+            return self.authenticate()
+
+        self.display_name = user_prefs["displayName"]
+        logger.debug("Display name is %s", self.display_name)
+
+        self.unit_system = user_prefs["measurementSystem"]
+        logger.debug("Unit system is %s", self.unit_system)
+
+        social_profile = self.__get_json(response.text, "VIEWER_SOCIAL_PROFILE")
+        self.full_name = social_profile["fullName"]
+        logger.debug("Fullname is %s", self.full_name)
+
+        if self.display_name == session_display_name:
+            return True
+        else:
+            logger.debug("Session not valid for user %s", self.display_name)
+            return self.authenticate()
+
+    def authenticate(self):
         """Login to Garmin Connect."""
 
         logger.debug("login: %s %s", self.username, self.password)
+        self.modern_rest_client.clear_cookies()
+        self.sso_rest_client.clear_cookies()
+
         get_headers = {"Referer": self.garmin_connect_login_url}
         params = {
             "service": self.modern_rest_client.url(),
@@ -284,6 +380,18 @@ class Garmin:
         self.full_name = social_profile["fullName"]
         logger.debug("Fullname is %s", self.full_name)
 
+        self.session_data = {
+            "display_name": self.display_name,
+            "session_cookies": requests.utils.dict_from_cookiejar(
+                self.modern_rest_client.get_cookies()
+            ),
+            "login_cookies": requests.utils.dict_from_cookiejar(
+                self.sso_rest_client.get_cookies()
+            ),
+        }
+
+        logger.debug("Cookies saved")
+
         return True
 
     def get_full_name(self):
@@ -297,16 +405,16 @@ class Garmin:
         return self.unit_system
 
     def get_stats(self, cdate: str) -> Dict[str, Any]:
-        """Return user activity summary for 'cdate' format 'YYYY-mm-dd' (compat for garminconnect)."""
+        """Return user activity summary for 'cdate' format 'YYYY-MM-DD' (compat for garminconnect)."""
 
         return self.get_user_summary(cdate)
 
     def get_user_summary(self, cdate: str) -> Dict[str, Any]:
-        """Return user activity summary for 'cdate' format 'YYYY-mm-dd'."""
+        """Return user activity summary for 'cdate' format 'YYYY-MM-DD'."""
 
         url = f"{self.garmin_connect_daily_summary_url}/{self.display_name}"
         params = {
-            "calendarDate": str(cdate),
+            "calendarDate": str(cdate)
         }
         logger.debug("Requesting user summary")
 
@@ -318,22 +426,22 @@ class Garmin:
         return response
 
     def get_steps_data(self, cdate):
-        """Fetch available steps data 'cDate' format 'YYYY-mm-dd'."""
+        """Fetch available steps data 'cDate' format 'YYYY-MM-DD'."""
 
         url = f"{self.garmin_connect_user_summary_chart}/{self.display_name}"
         params = {
-            "date": str(cdate),
+            "date": str(cdate)
         }
         logger.debug("Requesting steps data")
 
         return self.modern_rest_client.get(url, params=params).json()
 
-    def get_heart_rates(self, cdate):  #
-        """Fetch available heart rates data 'cDate' format 'YYYY-mm-dd'."""
+    def get_heart_rates(self, cdate):
+        """Fetch available heart rates data 'cDate' format 'YYYY-MM-DD'."""
 
         url = f"{self.garmin_connect_heartrates_daily_url}/{self.display_name}"
         params = {
-            "date": str(cdate),
+            "date": str(cdate)
         }
         logger.debug("Requesting heart rates")
 
@@ -348,7 +456,7 @@ class Garmin:
         }
 
     def get_body_composition(self, startdate: str, enddate=None) -> Dict[str, Any]:
-        """Return available body composition data for 'startdate' format 'YYYY-mm-dd' through enddate 'YYYY-mm-dd'."""
+        """Return available body composition data for 'startdate' format 'YYYY-MM-DD' through enddate 'YYYY-MM-DD'."""
 
         if enddate is None:
             enddate = startdate
@@ -359,15 +467,15 @@ class Garmin:
         return self.modern_rest_client.get(url, params=params).json()
 
     def get_max_metrics(self, cdate: str) -> Dict[str, Any]:
-        """Return available max metric data for 'cdate' format 'YYYY-mm-dd'."""
+        """Return available max metric data for 'cdate' format 'YYYY-MM-DD'."""
 
-        url = f"{self.garmin_connect_metrics_url}/{cdate}"
-        logger.debug("Requestng max metrics")
+        url = f"{self.garmin_connect_metrics_url}/{cdate}/{cdate}"
+        logger.debug("Requesting max metrics")
 
         return self.modern_rest_client.get(url).json()
 
     def get_hydration_data(self, cdate: str) -> Dict[str, Any]:
-        """Return available hydration data 'cdate' format 'YYYY-mm-dd'."""
+        """Return available hydration data 'cdate' format 'YYYY-MM-DD'."""
 
         url = f"{self.garmin_connect_daily_hydration_url}/{cdate}"
         logger.debug("Requesting hydration data")
@@ -375,7 +483,7 @@ class Garmin:
         return self.modern_rest_client.get(url).json()
 
     def get_respiration_data(self, cdate: str) -> Dict[str, Any]:
-        """Return available respiration data 'cdate' format 'YYYY-mm-dd'."""
+        """Return available respiration data 'cdate' format 'YYYY-MM-DD'."""
 
         url = f"{self.garmin_connect_daily_respiration_url}/{cdate}"
         logger.debug("Requesting respiration data")
@@ -383,7 +491,7 @@ class Garmin:
         return self.modern_rest_client.get(url).json()
 
     def get_spo2_data(self, cdate: str) -> Dict[str, Any]:
-        """Return available SpO2 data 'cdate' format 'YYYY-mm-dd'."""
+        """Return available SpO2 data 'cdate' format 'YYYY-MM-DD'."""
 
         url = f"{self.garmin_connect_daily_spo2_url}/{cdate}"
         logger.debug("Requesting SpO2 data")
@@ -398,25 +506,92 @@ class Garmin:
 
         return self.modern_rest_client.get(url).json()
 
+    def get_earned_badges(self) -> Dict[str, Any]:
+        """Return earned badges for current user."""
+
+        url = self.garmin_connect_earned_badges_url
+        logger.debug("Requesting earned badges for user")
+
+        return self.modern_rest_client.get(url).json()
+
+    def get_adhoc_challenges(self, start, limit) -> Dict[str, Any]:
+        """Return adhoc challenges for current user."""
+
+        url = self.garmin_connect_adhoc_challenges_url
+        params = {"start": str(start), "limit": str(limit)}
+        logger.debug("Requesting adhoc challenges for user")
+
+        return self.modern_rest_client.get(url, params=params).json()
+
+    def get_badge_challenges(self, start, limit) -> Dict[str, Any]:
+        """Return badge challenges for current user."""
+
+        url = self.garmin_connect_badge_challenges_url
+        params = {"start": str(start), "limit": str(limit)}
+        logger.debug("Requesting badge challenges for user")
+
+        return self.modern_rest_client.get(url, params=params).json()
+
+    def get_available_badge_challenges(self, start, limit) -> Dict[str, Any]:
+        """Return available badge challenges."""
+
+        url = self.garmin_connect_available_badge_challenges_url
+        params = {"start": str(start), "limit": str(limit)}
+        logger.debug("Requesting available badge challenges")
+
+        return self.modern_rest_client.get(url, params=params).json()
+
+    def get_non_completed_badge_challenges(self, start, limit) -> Dict[str, Any]:
+        """Return badge non-completed challenges for current user."""
+
+        url = self.garmin_connect_non_completed_badge_challenges_url
+        params = {"start": str(start), "limit": str(limit)}
+        logger.debug("Requesting badge challenges for user")
+
+        return self.modern_rest_client.get(url, params=params).json()
+
     def get_sleep_data(self, cdate: str) -> Dict[str, Any]:
         """Return sleep data for current user."""
 
-        url = f"{self.garmin_connect_sleep_daily_url}/{self.display_name}"
+        url = f"{self.garmin_connect_daily_sleep_url}/{self.display_name}"
         params = {"date": str(cdate), "nonSleepBufferMinutes": 60}
-
         logger.debug("Requesting sleep data")
 
         return self.modern_rest_client.get(url, params=params).json()
 
+    def get_stress_data(self, cdate: str) -> Dict[str, Any]:
+        """Return stress data for current user."""
+
+        url = f"{self.garmin_connect_daily_stress_url}/{cdate}"
+        logger.debug("Requesting stress data")
+
+        return self.modern_rest_client.get(url).json()
+
     def get_rhr_day(self, cdate: str) -> Dict[str, Any]:
         """Return resting heartrate data for current user."""
 
+        url = f"{self.garmin_connect_rhr_url}/{self.display_name}"
         params = {"fromDate": str(cdate), "untilDate": str(cdate), "metricId": 60}
-        url = f"{self.garmin_connect_rhr}/{self.display_name}"
         logger.debug("Requesting resting heartrate data")
 
         return self.modern_rest_client.get(url, params=params).json()
 
+    def get_training_readiness(self, cdate: str) -> Dict[str, Any]:
+        """Return training readiness data for current user."""
+
+        url = f"{self.garmin_connect_training_readiness_url}/{cdate}"
+        logger.debug("Requesting training readiness data")
+
+        return self.modern_rest_client.get(url).json()
+
+    def get_training_status(self, cdate: str) -> Dict[str, Any]:
+        """Return training status data for current user."""
+
+        url = f"{self.garmin_connect_training_status_url}/{cdate}"
+        logger.debug("Requesting training status data")
+
+        return self.modern_rest_client.get(url).json()
+
     def get_devices(self) -> Dict[str, Any]:
         """Return available devices for the current user account."""
 
@@ -465,12 +640,58 @@ class Garmin:
     def get_last_activity(self):
         """Return last activity."""
 
-        activities = self.get_activities(0,1)
+        activities = self.get_activities(0, 1)
         if activities:
             return activities[-1]
 
         return None
 
+    def upload_fit_activity(self, fit_file):
+        """Upload activity in fit format from file."""
+
+        logger.debug("Uploading activity in fit format from file")
+        with open(fit_file, 'rb') as file:
+            return self.modern_rest_client.post(self.garmin_connect_fit_upload_url, {}, {}, file)
+
+    def get_activities_by_date(self, startdate, enddate, activitytype=None):
+        """
+        Fetch available activities between specific dates
+        :param startdate: String in the format YYYY-MM-DD
+        :param enddate: String in the format YYYY-MM-DD
+        :param activitytype: (Optional) Type of activity you are searching
+                             Possible values are [cycling, running, swimming,
+                             multi_sport, fitness_equipment, hiking, walking, other]
+        :return: list of JSON activities
+        """
+
+        activities = []
+        start = 0
+        limit = 20
+        # mimicking the behavior of the web interface that fetches 20 activities at a time
+        # and automatically loads more on scroll
+        url = self.garmin_connect_activities
+        params = {
+            "startDate": str(startdate),
+            "endDate": str(enddate),
+            "start": str(start),
+            "limit": str(limit),
+        }
+        if activitytype:
+            params["activityType"] = str(activitytype)
+
+        logger.debug(f"Requesting activities by date from {startdate} to {enddate}")
+        while True:
+            params["start"] = str(start)
+            logger.debug(f"Requesting activities {start} to {start+limit}")
+            act = self.modern_rest_client.get(url, params=params).json()
+            if act:
+                activities.extend(act)
+                start = start + limit
+            else:
+                break
+
+        return activities
+
     class ActivityDownloadFormat(Enum):
         """Activitie variables."""
 
@@ -538,6 +759,15 @@ class Garmin:
 
         return self.modern_rest_client.get(url).json()
 
+    def get_activity_evaluation(self, activity_id):
+        """Return activity self evaluation details."""
+
+        activity_id = str(activity_id)
+        url = f"{self.garmin_connect_activity}/{activity_id}"
+        logger.debug("Requesting self evaluation data for activity id %s", activity_id)
+
+        return self.modern_rest_client.get(url).json()
+
     def get_activity_details(self, activity_id, maxchart=2000, maxpoly=4000):
         """Return activity details."""
 
